/// Wraps [`format_args`] casting all arguments to `AttValueSafe`.  The return
/// value is wrapped to also implement `AttValueSafe`.  Literal `"` and `&`
/// characters in the format string must be escaped as `&quot;` and `&amp;`.
#[macro_export]
macro_rules! format_att_value {
    ($format:expr $(,$args:expr)*) => {
        {
            // We are defining a new struct in here because otherwise it could
            // be used to circumvent the trait (we cannot just define it in
            // crate::wrappers with a private constructor since macros are
            // expanded in the calling code, where they can't call private
            // constructors defined in this crate).
            struct FormattedAttValue<'a>(std::fmt::Arguments<'a>);
            impl $crate::AttValueSafe for FormattedAttValue<'_> {}
            impl std::fmt::Display for FormattedAttValue<'_> {
                fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                    self.0.fmt(f)
                }
            }
            FormattedAttValue(
                format_args!($format $(,&$args as &dyn $crate::AttValueSafe)*)
            )
        }
    };
}

/// Wraps [`format_args`] casting all arguments to `PcdataSafe`.  The return
/// value is wrapped to also implement `PcdataSafe`. Literal `<` and `&`
/// characters in the format string must be escaped as `&lt;` and `&amp;`
/// respectively. Do **NOT** use this to build tags (instead use the secure
/// methods from [`XmlWriter`](crate::XmlWriter)).
#[macro_export]
macro_rules! format_text {
    ($format:expr $(,$args:expr)*) => {
        {
            // See the comment in the previous macro.
            struct FormattedPcdata<'a>(std::fmt::Arguments<'a>);
            impl $crate::PcdataSafe for FormattedPcdata<'_> {}
            impl std::fmt::Display for FormattedPcdata<'_> {
                fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                    self.0.fmt(f)
                }
            }
            FormattedPcdata(
                format_args!($format $(,&$args as &dyn $crate::PcdataSafe)*)
            )
        }
    };
}

/// Writes an XML tag to a given `XmlWriter` and takes care of closing the tag.
///
/// ## Basic usage example
/// ```
/// # use std::fmt::Error;
/// # use xmlsafe::{XmlWriter, format_text, escape_text, tag};
/// #
/// struct User {name: String, id: u8}
///
/// fn list_users(mut w: XmlWriter, users: Vec<User>) -> Result<(), Error> {
///     tag!(w, "ul", {
///         for user in users {
///             tag!(w, "li", "id"=user.id, {
///                 w.write(escape_text(user.name))?;
///             });
///         }
///     });
///     Ok(())
/// }
///
/// let mut out = String::new();
/// let users = vec![User{name: "Alice".into(), id: 3}, User{name: "Bob".into(), id: 5}];
/// list_users(XmlWriter::new(&mut out), users).unwrap();
/// assert_eq!(out, "<ul><li id=\"3\">Alice</li><li id=\"5\">Bob</li></ul>");
/// ```
///
/// ## Streaming attribute values
///
/// When generating attribute values you can avoid allocations by using an
/// [`AttValueWriter`](crate::AttValueWriter), for example:
///
/// ```
/// # use std::fmt::Error;
/// # use xmlsafe::{XmlWriter, tag};
/// #
/// fn generate(mut w: XmlWriter, max: u8) -> Result<(), Error> {
///     tag!(w, "example", "data" << |value| {
///         for i in 1..max + 1 {
///             value.write(i)?;
///             value.write(" ")?;
///         }
///         Ok(())
///     });
///     Ok(())
/// }
///
/// let mut out = String::new();
/// generate(XmlWriter::new(&mut out), 5).unwrap();
/// assert_eq!(out, "<example data=\"1 2 3 4 5 \"/>");
/// ```
///
/// ## Tag constructors
///
/// The first argument supplied to the `tag!` macro can actually be any expression
/// that returns a type implementing the [`TagConstructor`](crate::TagConstructor) trait.
/// A tag constructor specifies the tag name and an initial set of attributes.
/// This lets you easily introduce additional type safety and convenience:
/// prevent typos, enable autocompletion and construct attributes from
/// higher-level types, all without compromising extensibility.  Further attributes
/// and content can be added via the `tag!` macro as usual. For example:
///
/// ```
/// # use xmlsafe::{AttWriter, NameSafe, TagConstructor, XmlWriter, tag};
/// # use std::fmt::Error;
/// #
/// struct Circle { center: Point, radius: f64 }
/// struct Point { x: f64, y: f64 }
///
/// impl TagConstructor for Circle {
///     fn tag_name(&self) -> &dyn NameSafe {
///         &"circle"
///     }
///
///     fn attributes<'a>(&self, writer: AttWriter<'a>) -> Result<AttWriter<'a>, Error> {
///         writer
///             .attr("cx", self.center.x)?
///             .attr("cy", self.center.y)?
///             .attr("r", self.radius)
///     }
/// }
///
/// fn generate(mut w: XmlWriter) -> Result<(), Error> {
///     tag!(
///         w,
///         Circle {
///             center: Point{ x: 3.0, y: 4.0 },
///             radius: 5.0,
///         },
///         "class" = "my-rectangle"
///     );
///     Ok(())
/// }
///
/// let mut out = String::new();
/// generate(XmlWriter::new(&mut out)).unwrap();
/// assert_eq!(out, r#"<circle cx="3" cy="4" r="5" class="my-rectangle"/>"#);
/// ```
#[macro_export]
macro_rules! tag {
    // This macro should play nice with rustfmt. Therefore :block is used for
    // code that should be indented and the attributes are separated by commas.
    // This macro should also work well with rust-analyzer, which is why
    // the macro is split up into two rules instead of being more DRY.
    ($writer:ident, $constructor:expr $(, $k:literal$t:tt$v:expr)*) => {
        {
            let constructor: &dyn $crate::TagConstructor = &$constructor;
            #[allow(unused_mut)]
            let mut att_writer = constructor.attributes($writer.open_start_tag(constructor.tag_name())?)?;

            #[allow(unused_macros)]
            macro_rules! attr {
                ($ak:literal << $av:expr) => {
                    let mut att_val_writer = att_writer.key($ak)?;
                    let fun: &dyn Fn(&mut $crate::AttValueWriter) -> Result<(), std::fmt::Error> = &$av;
                    fun(&mut att_val_writer)?;
                    att_writer = att_val_writer.end()?;
                };
                ($ak:literal=$av:expr) => {
                    att_writer.write_attr($ak, $av)?;
                };
            }
            $(attr!($k$t$v);)*

            #[allow(unused_assignments)]
            {
                $writer = att_writer.close_empty()?;
            }
        };
    };
    ($writer:ident, $constructor:expr $(, $k:literal$t:tt$v:expr)*, $content:block) => {
        {
            let constructor: &dyn $crate::TagConstructor = &$constructor;
            #[allow(unused_mut)]
            let mut att_writer = constructor.attributes($writer.open_start_tag(constructor.tag_name())?)?;

            #[allow(unused_macros)]
            macro_rules! attr {
                ($ak:literal << $av:expr) => {
                    let mut att_val_writer = att_writer.key($ak)?;
                    let fun: &dyn Fn(&mut $crate::AttValueWriter) -> Result<(), std::fmt::Error> = &$av;
                    fun(&mut att_val_writer)?;
                    att_writer = att_val_writer.end()?;
                };
                ($ak:literal=$av:expr) => {
                    att_writer.write_attr($ak, $av)?;
                };
            }
            $(attr!($k$t$v);)*

            $writer = att_writer.close()?;
            $content
            $writer.write_end_tag(constructor.tag_name())?;
        };
    };
}

#[cfg(test)]
mod tests {
    use std::fmt::Error;

    use crate::{AttValueWriter, XmlWriter};

    #[test]
    fn test_empty_tag() {
        fn generate(mut w: XmlWriter) -> Result<(), std::fmt::Error> {
            tag!(w, "example");
            tag!(w, "example", "x" = 33);
            Ok(())
        }
        let mut out = String::new();
        generate(XmlWriter::new(&mut out)).unwrap();
        assert_eq!(out, "<example/><example x=\"33\"/>");
    }

    #[test]
    fn test_stream_att_value_with_function_pointer() {
        fn gen_value(value: &mut AttValueWriter) -> Result<(), Error> {
            value.write("from function")?;
            Ok(())
        }
        fn generate(mut w: XmlWriter) -> Result<(), Error> {
            tag!(w, "example", "key" << gen_value);
            Ok(())
        }
        let mut out = String::new();
        generate(XmlWriter::new(&mut out)).unwrap();
        assert_eq!(out, "<example key=\"from function\"/>");
    }
}
